@@ -27,7 +27,7 @@ import warnings
 from pkgutil import find_loader
 from collections import OrderedDict
 from copy import copy, deepcopy
-from typing import Union  # noqa F401
+from typing import Union, Optional  # noqa F401
 
 import numpy as np
 from osgeo import gdal, gdal_array, gdalnumeric
@@ -64,9 +64,9 @@ class GeoArray(object):
     some functions for quickly visualizing the data as a map, a simple image or an interactive image.
     """
 
-    def __init__(self, path_or_array, geotransform=None, projection=None, bandnames=None, nodata=None, progress=True,
-                 q=False):
-        # type: (Union[str, np.ndarray, GeoArray], tuple, str, list, float, bool, bool) -> None
+    def __init__(self, path_or_array, geotransform=None, projection=None, bandnames=None, nodata=None, basename='',
+                 progress=True, q=False):
+        # type: (Union[str, np.ndarray, GeoArray], tuple, str, list, float, str, bool, bool) -> None
         """Get an instance of GeoArray.
 
         :param path_or_array:   a numpy.ndarray or a valid file path
@@ -76,6 +76,7 @@ class GeoArray(object):
         :param bandnames:       names of the bands within the input array, e.g. ['mask_1bit', 'mask_clouds'],
                                 (default: ['B1', 'B2', 'B3', ...])
         :param nodata:          nodata value
+        :param basename:        a short base name of the dataset (e.g., used in some status messages)
         :param progress:        show progress bars (default: True)
         :param q:               quiet mode (default: False)
         """
@@ -90,7 +91,7 @@ class GeoArray(object):
         if isinstance(path_or_array, str):
             assert ' ' not in path_or_array, "The given path contains whitespaces. This is not supported by GDAL."
 
-            if not os.path.exists(path_or_array):
+            if not os.path.exists(path_or_array) and not path_or_array.startswith('/vsi'):
                 raise FileNotFoundError(path_or_array)
 
         if isinstance(path_or_array, GeoArray) or issubclass(getattr(path_or_array, '__class__'), GeoArray):
@@ -99,16 +100,18 @@ class GeoArray(object):
             self.geotransform = geotransform or self.geotransform
             self.projection = projection or self.projection
             self.bandnames = bandnames or list(self.bandnames.keys())
+            self.basename = basename or self.basename
             self._nodata = nodata if nodata is not None else self._nodata
             self.progress = False if progress is False else self.progress
-            self.q = q if q is not None else self.q
+            self.q = q or self.q
 
         else:
             self._initParams = dict([x for x in locals().items() if x[0] != "self"])
             self.arg = path_or_array
             self._arr = path_or_array if isinstance(path_or_array, np.ndarray) else None
             self.filePath = path_or_array if isinstance(path_or_array, str) and path_or_array else None
-            self.basename = os.path.splitext(os.path.basename(self.filePath))[0] if not self.is_inmem else 'IN_MEM'
+            basename_default = os.path.splitext(os.path.basename(self.filePath))[0] if not self.is_inmem else 'IN_MEM'
+            self.basename = basename or basename_default
             self.progress = progress
             self.q = q
             self._arr_cache = None  # dict containing key 'pos' and 'arr_cached'
@@ -379,7 +382,14 @@ class GeoArray(object):
 
     @nodata.setter
     def nodata(self, value):
-        # type: (Union[int, None]) -> None
+        # type: (Optional[bool, int, float]) -> None
+        if isinstance(value, np.bool_):
+            value = bool(value)
+        elif isinstance(value, np.integer):
+            value = int(value)
+        elif isinstance(value, np.floating):
+            value = float(value)
+
         self._nodata = value
 
         if self._metadata and value is not None:
@@ -991,65 +1001,51 @@ class GeoArray(object):
 
         driver = gdal.GetDriverByName(fmt)
         if driver is None:
-            raise Exception("'%s' is not a supported GDAL driver. Refer to www.gdal.org/formats_list.html for full "
-                            "list of GDAL driver codes." % fmt)
+            raise Exception("'%s' is not a supported GDAL driver. Refer to https://gdal.org/drivers/raster/index.html "
+                            "for full list of GDAL driver codes." % fmt)
 
         if not os.path.isdir(os.path.dirname(out_path)):
             os.makedirs(os.path.dirname(out_path))
 
         envi_metadict = self.metadata.to_ENVI_metadict()
 
-        #####################
-        # write raster data #
-        #####################
+        ###########################
+        # get source GDAL dataset #
+        ###########################
+
+        ds_src: gdal.Dataset
+        ds_out: gdal.Dataset
 
         if self.is_inmem:
-            # write in-memory data
-            ds_inmem = get_GDAL_ds_inmem(self.arr,  # expects rows,columns,bands
-                                         self.geotransform, self.projection,
-                                         self._nodata)  # avoid to compute the nodata value here, so use private attrib.
-
-            # write dataset
-            ds_out = driver.CreateCopy(out_path, ds_inmem,
-                                       options=creationOptions or [])  # noqa
-
-            # # rows, columns, bands => bands, rows, columns
-            # out_arr = self.arr if self.ndim == 2 else np.swapaxes(np.swapaxes(self.arr, 0, 2), 1, 2)
-            # gdalnumeric.SaveArray(out_arr, out_path, format=fmt, prototype=ds_inmem)  # expects bands,rows,columns
-            # ds_out = gdal.Open(out_path)
-
-            del ds_inmem
-            del ds_out
+            ds_src = get_GDAL_ds_inmem(self.arr,  # expects rows,columns,bands
+                                       self.geotransform, self.projection,
+                                       self._nodata)  # avoid to compute the nodata value here, so use private attrib.
 
         else:
-            # transform linked image dataset (not in-memory) to target format
-            src_ds = gdal.Open(self.filePath)
-            if not src_ds:
-                raise Exception('Error reading file:  ' + gdal.GetLastErrorMsg())
-
+            ds_src = gdal.Open(self.filePath)
             # metadomains = {dom: src_ds.GetMetadata(dom) for dom in src_ds.GetMetadataDomainList()}
 
-            gdal.Translate(out_path, src_ds, format=fmt, creationOptions=creationOptions)
-            del src_ds
+        if not ds_src:
+            raise Exception('Error reading file:  ' + gdal.GetLastErrorMsg())
 
-        if not os.path.exists(out_path):
-            raise Exception(gdal.GetLastErrorMsg())
-
-        ################
-        # set metadata #
-        ################
-
-        # NOTE:  The dataset has to be written BEFORE metadata are added. Otherwise, metadata are not written.
-
-        ds_out = gdal.Open(out_path, gdal.GA_Update)
+        #########################################
+        # write output dataset and set metadata #
+        #########################################
 
         try:
-            if not ds_out:
-                raise Exception('Error reading file:  ' + gdal.GetLastErrorMsg())
+            # disable to write separate metadata XML files
+            os.environ['GDAL_PAM_ENABLED'] = 'NO'
 
             # ENVI #
             ########
             if fmt == 'ENVI':
+                # NOTE:  The dataset has to be written BEFORE metadata are added. Otherwise, metadata are not written.
+
+                # write ds_src to disk and re-open it to add the metadata
+                gdal.Translate(out_path, ds_src, format=fmt, creationOptions=creationOptions)
+                del ds_src
+                ds_out = gdal.Open(out_path, gdal.GA_Update)
+
                 for bidx in range(self.bands):
                     band = ds_out.GetRasterBand(bidx + 1)
 
@@ -1058,24 +1054,22 @@ class GeoArray(object):
                         band.SetDescription(bandname)
                         assert band.GetDescription() == bandname
 
-                    # NOTE: band-specific nodata values are not supported by the ENVI header format
-                    if 'nodata' in envi_metadict:
-                        nodataVal = self.metadata.band_meta['nodata'][bidx]
-                        band.SetNoDataValue(nodataVal)
-                        assert band.GetNoDataValue() == nodataVal
-
                     del band
 
                 # avoid that band names are written to global meta
                 if 'band_names' in envi_metadict:
                     del envi_metadict['band_names']
 
+                # the expected key name is 'data_ignore_value', see below
                 if 'nodata' in envi_metadict:
                     del envi_metadict['nodata']
 
                 # set data_ignore_value in case self.metadata.band_meta contains a unique nodata value
-                if 'nodata' in self.metadata.band_meta and len(set(self.metadata.band_meta['nodata'])) == 1:
-                    envi_metadict['data_ignore_value'] = str(self.metadata.band_meta['nodata'][0])
+                if 'nodata' in self.metadata.band_meta:
+                    if len(set(self.metadata.band_meta['nodata'])) == 1:
+                        envi_metadict['data_ignore_value'] = str(self.metadata.band_meta['nodata'][0])
+                    else:
+                        warnings.warn("Band-specific nodata values are not supported by the ENVI header format.")
 
                 ds_out.SetMetadata(envi_metadict, 'ENVI')
 
@@ -1083,44 +1077,60 @@ class GeoArray(object):
                     ds_out.SetDescription(envi_metadict['description'])
 
                 ds_out.FlushCache()
-                gdal.Unlink(out_path + '.aux.xml')
+                # gdal.Unlink(out_path + '.aux.xml')
 
-            elif self.metadata.all_meta:
-                # set global domain metadata
-                if self.metadata.global_meta:
-                    ds_out.SetMetadata(dict((k, repr(v)) for k, v in self.metadata.global_meta.items()))
+            else:
+                ds_out = ds_src
 
-                if 'description' in envi_metadict:
-                    ds_out.SetDescription(envi_metadict['description'])
+                # set metadata
+                if self.metadata.all_meta:
 
-                # set band domain metadata
-                bandmeta_dict = self.metadata.to_DataFrame().astype(str).to_dict()
+                    # set global domain metadata
+                    if self.metadata.global_meta:
+                        ds_out.SetMetadata(dict((k, repr(v)) for k, v in self.metadata.global_meta.items()))
 
-                for bidx in range(self.bands):
-                    band = ds_out.GetRasterBand(bidx + 1)
-                    bandmeta = bandmeta_dict[bidx].copy()
+                    if 'description' in envi_metadict:
+                        ds_out.SetDescription(envi_metadict['description'])
 
-                    # filter global metadata out
-                    bandmeta = {k: v for k, v in bandmeta.items() if k not in self.metadata.global_meta}
-                    # meta2write = dict((k, repr(v)) for k, v in self.metadata.band_meta.items() if v is not np.nan)
+                    # set band domain metadata
+                    bandmeta_dict = self.metadata.to_DataFrame().astype(str).to_dict()
 
-                    if 'band_names' in bandmeta:
-                        band.SetDescription(self.metadata.band_meta['band_names'][bidx].strip())
-                        del bandmeta['band_names']
+                    for bidx in range(self.bands):
+                        band = ds_out.GetRasterBand(bidx + 1)
+                        bandmeta = bandmeta_dict[bidx].copy()
 
-                    if 'nodata' in bandmeta:
-                        band.SetNoDataValue(self.metadata.band_meta['nodata'][bidx])
-                        del bandmeta['nodata']
+                        # filter global metadata out
+                        bandmeta = {k: v for k, v in bandmeta.items() if k not in self.metadata.global_meta}
+                        # meta2write = dict((k, repr(v)) for k, v in self.metadata.band_meta.items() if v is not np.nan)
 
-                    if bandmeta:
-                        band.SetMetadata(bandmeta)
+                        if 'band_names' in bandmeta:
+                            band.SetDescription(self.metadata.band_meta['band_names'][bidx].strip())
+                            del bandmeta['band_names']
 
-                    band.FlushCache()
-                    del band
+                        if 'nodata' in bandmeta:
+                            band.SetNoDataValue(self.metadata.band_meta['nodata'][bidx])
+                            del bandmeta['nodata']
 
-                ds_out.FlushCache()
+                        if bandmeta:
+                            band.SetMetadata(bandmeta)
+
+                        band.FlushCache()
+                        del band
+
+                    ds_out.FlushCache()
+
+                # write ds_out to disk,
+                # -> writes the in-memory array or transforms the linked dataset into the target format
+                gdal.Translate(out_path, ds_out, format=fmt, creationOptions=creationOptions)
+                del ds_src
 
         finally:
+            if 'GDAL_PAM_ENABLED' in os.environ:
+                del os.environ['GDAL_PAM_ENABLED']
+
+            if not os.path.exists(out_path):
+                raise Exception(gdal.GetLastErrorMsg())
+
             del ds_out
 
     def dump(self, out_path):
@@ -1150,6 +1160,8 @@ class GeoArray(object):
         out_nodata = in_nodata if in_nodata is not None else -9999
         if not np.can_cast(out_nodata, image2plot.dtype):
             image2plot = image2plot.astype(np.int32)
+        if image2plot.dtype == bool:
+            image2plot = image2plot.astype(int)
 
         # rotated images always have to be resampled for plotting
         if not ignore_rotation and (gt[2] or gt[4]):
@@ -1301,7 +1313,7 @@ class GeoArray(object):
                 ax = plt.gca()
 
             rows, cols = image2plot.shape[:2]
-            im = ax.imshow(image2plot.astype(int),
+            im = ax.imshow(image2plot,
                            palette,
                            interpolation=interpolation,
                            extent=(0, cols, rows, 0),
@@ -1732,10 +1744,10 @@ class GeoArray(object):
 
         # apply zslice to bandnames and metadata
         if zslicing:
-            bNs_out = list(np.array(list(self.bandnames))[zslice])
+            bNs_out = list(np.array(list(self._bandnames))[zslice]) if self._bandnames else None
             _meta_out = self.metadata.get_subset(bands2extract=zslice)
         else:
-            bNs_out = list(self.bandnames)
+            bNs_out = list(self._bandnames) if self._bandnames else None
             _meta_out = self.meta
 
         sub_gA.gt = sub_gt
